#include "config.h"

#include <sys/time.h>

#include <set>

#include "general/AtExit.hh"
#include "general/gpstime.hh"
#include "general/mutexlock.hh"
#include "general/Task.hh"
#include "general/TaskThread.hh"
#include "general/ThreadPool.hh"

#include "diskcache.hh"
#include "dirinfo.hh"
#include "HotDirectory.hh"

namespace
{
  using General::GPSTime;
  using diskCache::DirectoryInfo;
  using diskCache::mtime_type;
  using GenericAPI::queueLogEntry;
  using diskCache::LogMessage;

  static const std::string JOB_INFO = "HOT_DIR";

  //---------------------------------------------------------------------
  /// \brief Process the collection of hot directories.
  ///
  /// This maintains the thread that is responsible for scanning
  /// directories that are being updated later than the user
  /// specified lower bound time.
  //---------------------------------------------------------------------
  class daemon
    : General::Task
  {
  public:
    //-------------------------------------------------------------------
    /// \brief Storage type for directory information.
    ///
    /// Having the directory information stored in a shared pointer
    /// container increases thread safety and diminishes the needs
    /// for locks.
    //-------------------------------------------------------------------
    typedef diskCache::HotDirectory::weak_dir_info_type	weak_dir_info_type;

    //-------------------------------------------------------------------
    /// \brief Collection of patterns used to exclude directories
    ///
    /// This information needs to be preserved for doing a scan
    /// of a directory.
    /// It should be treated as an opaque object.
    //-------------------------------------------------------------------
    typedef diskCache::HotDirectory::dir_container_type dir_container_type;

    //-------------------------------------------------------------------
    /// \brief Storage class of state information.
    ///
    /// This information is relevant for each directory that is 
    /// considered active.
    //-------------------------------------------------------------------
    struct info {
      //-----------------------------------------------------------------
      /// \brief Reference to directory information
      ///
      /// This reference is used to request the scanning.
      //-----------------------------------------------------------------
      weak_dir_info_type		m_dir_info_ref;
      //-----------------------------------------------------------------
      /// \brief Collection of patterns used to exclude directories
      ///
      /// This information needs to be preserved for doing a scan
      /// of a directory.
      /// It should be treated as an opaque object.
      //-----------------------------------------------------------------
      dir_container_type	m_exclude_dirs;

      //-----------------------------------------------------------------
      /// \brief Default Constructor
      //-----------------------------------------------------------------
      info( );

      //-----------------------------------------------------------------
      /// \brief Constructor
      //-----------------------------------------------------------------
      info( weak_dir_info_type DirInfo,
	    const dir_container_type& ExcludeDirs );
    };

    //-------------------------------------------------------------------
    /// \brief Default constructor
    //-------------------------------------------------------------------
    daemon( );

    //-------------------------------------------------------------------
    /// \brief Start the daemon thread
    //-------------------------------------------------------------------
    static void Start( );

    //-------------------------------------------------------------------
    /// \brief Update each of entries in the collection of hot directories.
    //-------------------------------------------------------------------
    virtual void operator()( );

  private:
    class no_next
      : public std::out_of_range
    {
    public:
      no_next( )
	: std::out_of_range( "Hot Directory List exhausted" )
      {
      }
    };

    //-------------------------------------------------------------------
    /// \brief Get the next hot directory to scan.
    ///
    /// \param[in] DirInfoRef
    /// \param[in] ExcludeDirs
    ///
    /// \return
    ///     The reference that represents 
    //-------------------------------------------------------------------
    static weak_dir_info_type
    next( weak_dir_info_type DirInfoRef,
	  dir_container_type& ExcludeDirs );

  };

  //---------------------------------------------------------------------
  /// \brief Container type for storing hot directory information.
  ///
  /// This container class stores information about directories
  /// that are considered hot.
  /// The container is indexed by the directory path.
  //---------------------------------------------------------------------
  typedef std::map< std::string,
		    daemon::info > hot_directory_container_type;

  //---------------------------------------------------------------------
  /// \brief Obtain lock
  ///
  /// The use of a function ensures the proper initialization of the data
  /// without having to depend on the link order initialization.
  //---------------------------------------------------------------------
  inline MutexLock::lock_type&
  baton( )
  {
    static MutexLock::lock_type	baton =
      MutexLock::Initialize( );
    
    return baton;
  }

  //---------------------------------------------------------------------
  /// \brief Obtain list of registered hot directories
  ///
  /// The use of a function ensures the proper initialization of the data
  /// without having to depend on the link order initialization.
  //---------------------------------------------------------------------
  inline hot_directory_container_type&
  hot_list( )
  {
    static hot_directory_container_type retval;

    return retval;
  }

  //---------------------------------------------------------------------
  /// \brief Register Entry in collection of hot directories
  ///
  /// This method registers Entry with the collection of hot
  /// directories only if it is not already a part of the collection.
  ///
  /// \param[in] Entry
  ///     The directory entry to be registered.
  /// \param[in] ExcludeDirs
  ///     A list of known directory names that should not be scanned.
  //---------------------------------------------------------------------
  inline void
  cond_register( daemon::weak_dir_info_type Entry,
		 const daemon::dir_container_type& ExcludeDirs )
  {
    General::SharedPtr< DirectoryInfo >	dir( Entry.lock( ) );
    General::SharedPtr< DirectoryInfo >	hl_dir;

    if ( ! dir )
    {
      // Cannot register a directory that has already been released.
      return;
    }

    MutexLock	l( baton( ) );

    hot_directory_container_type::iterator
      pos = hot_list( ).find( dir->GetPath( ).Var( ) );
    if ( pos != hot_list( ).end( ) )
    {
      hl_dir = pos->second.m_dir_info_ref.lock( );

      if (  hl_dir == dir )
      {
	//---------------------------------------------------------------
	// Directory has already been registered.
	//---------------------------------------------------------------
	return;
      }
      //-----------------------------------------------------------------
      // Reset the directory
      //-----------------------------------------------------------------
      pos->second.m_dir_info_ref = Entry;
      return;
    }
    //-------------------------------------------------------------------
    // Create new entry.
    //-------------------------------------------------------------------
    hot_directory_container_type::value_type
      n( dir->GetPath( ).Var( ),
	 hot_directory_container_type::mapped_type( Entry,
						    ExcludeDirs ) );
    hot_list( ).insert( n );

    std::ostringstream	msg;

    msg << "Added " << dir->getPath( )
	<< " to list of hot directories."
      ;
    queueLogEntry( msg.str( ),
		   GenericAPI::LogEntryGroup_type::MT_NOTE,
		   0,
		   "cond_register@HotDirectory.cc",
		   JOB_INFO );

    //-------------------------------------------------------------------
    // Ensure the scan daemon is running
    //-------------------------------------------------------------------
    daemon::Start( );
  }

  //---------------------------------------------------------------------
  /// \brief Check if Entry is hot
  ///
  /// This method will check if Entry has been modified within the
  /// specified time constraint.
  /// If it has, then it is considered hot.
  ///
  /// \param[in] Entry
  ///     The directory entry to be checked.
  ///
  /// \return
  ///     The value true is returned if the Entry is considered hot,
  ///     false otherwise.
  //---------------------------------------------------------------------
  inline bool
  is_hot( DirectoryInfo* Entry )
  {
    struct timeval timev;
    gettimeofday( &timev, NULL );

    const mtime_type localtime_secs = timev.tv_sec;
    
    return ( ( localtime_secs - Entry->TimeModified( ) )
	     < diskCache::HotDirectory::HotLowerBound( ) );
  }

  inline bool
  is_hot( daemon::weak_dir_info_type Entry )
  {
    General::SharedPtr< DirectoryInfo >
      dir( Entry.lock( ) );
    if ( ! dir )
    {
      return false;
    }
    return is_hot( dir.get( ) );
  }

  //---------------------------------------------------------------------
  /// \brief Check if Entry is already registed as a hot directory.
  ///
  /// Checks to see if Entry is alread a member of the collection
  /// of hot directories.
  ///
  /// \param[in] Entry
  ///     The directory entry to be considered.
  ///
  /// \return
  ///     If Entry is already a member of the collection of hot
  ///     directories, then a value of true is returned,
  ///     false otherwise.
  //---------------------------------------------------------------------
  inline bool
  is_registered( daemon::weak_dir_info_type Entry )
  {
    General::SharedPtr< DirectoryInfo >	dir( Entry.lock( ) );

    if ( ! dir )
    {
      //-----------------------------------------------------------------
      // Any empty directory is considered registered.
      //-----------------------------------------------------------------
      return true;
    }

    MutexLock	l( baton( ) );

    hot_directory_container_type::const_iterator
      pos = hot_list( ).find( dir->GetPath( ).Var( ) );
    return ( ( pos != hot_list( ).end( ) )
	     && ( pos->second.m_dir_info_ref.lock( ).get( ) == dir.get( ) ) );
  }

  //---------------------------------------------------------------------
  /// \brief Remove Entry from collection of hot directories.
  ///
  /// Rmeoves Entry from the collection of hot directories.
  /// This operation needs to be very careful as the requested
  /// Entry for removal may currently be in a scanning mode.
  /// It also needs to be careful as the Entry may be in
  /// its destructor state at the time of this call.
  ///
  /// \param[in] Entry
  ///     The directory entry to be removed from the collection
  ///     of hot directories.
  //---------------------------------------------------------------------
  inline void
  unregister( daemon::weak_dir_info_type Entry )
  {
    General::SharedPtr< DirectoryInfo >	dir( Entry.lock( ) );

    if ( ! dir )
    {
      //-----------------------------------------------------------------
      // No info is available.
      //-----------------------------------------------------------------
      return;
    }

    MutexLock	l( baton( ) );

    hot_directory_container_type::iterator
      pos = hot_list( ).find( dir->GetPath( ).Var( ) );

    if ( ( pos != hot_list( ).end( ) )
	 && ( pos->second.m_dir_info_ref.lock( ).get( ) == dir.get( ) ) )
    {
      std::ostringstream	msg;

      msg << "Removed " << pos->first
	  << " from the list of hot directories."
	;
      queueLogEntry( msg.str( ),
		     GenericAPI::LogEntryGroup_type::MT_NOTE,
		     0,
		     "daemon::unregister()@HotDirectory.cc",
		     JOB_INFO );
      hot_list( ).erase( pos );
    }
  }

  //---------------------------------------------------------------------
  /// Create a new instance of the object being careful to properly
  /// initialize data members.
  //---------------------------------------------------------------------
  daemon::info::
  info( )
  {
  }

  //---------------------------------------------------------------------
  /// Construct an instance of the object with all members of the object
  /// specified.
  //---------------------------------------------------------------------
  daemon::info::
  info( weak_dir_info_type DirInfoRef,
	const dir_container_type& ExcludeDirs )
    : m_dir_info_ref( DirInfoRef ),
      m_exclude_dirs( ExcludeDirs )
  {
  }

  //---------------------------------------------------------------------
  /// Ensure proper initialization of task.
  //---------------------------------------------------------------------
  daemon::
  daemon( )
  {
    delete_on_completion( true );
  }

  //---------------------------------------------------------------------
  /// Obtain the next logical directory to scan.
  /// This routine will skip any directory that is only referenced
  /// within this list.
  //---------------------------------------------------------------------
  daemon::weak_dir_info_type daemon::
  next( weak_dir_info_type DirInfoRef,
	dir_container_type& ExcludeDirs )
  { 
    MutexLock l( baton( ) );

    hot_directory_container_type::iterator pos;

    General::SharedPtr< DirectoryInfo >	dir( DirInfoRef.lock( ) );

    //-------------------------------------------------------------------
    // Obtain the next logical entry
    //-------------------------------------------------------------------
    if ( dir )
    {
      //-----------------------------------------------------------------
      // Since this is NOT the first time called, get the node just
      // beyond the current postion.
      //-----------------------------------------------------------------
      pos = hot_list( ).upper_bound( dir->GetPath( ).Var( ) );
    } else {
      //-----------------------------------------------------------------
      // As this is the first time that it is called, start at the
      // beginning
      //-----------------------------------------------------------------
      pos = hot_list( ).begin( );
    }
    //-------------------------------------------------------------------
    // Loop over the list to skip nodes that should be ignored.
    //-------------------------------------------------------------------
    while( pos != hot_list( ).end( )
	   && ( pos->second.m_dir_info_ref.expired( ) == true ) )
    {
      //-----------------------------------------------------------------
      // Skip past any node that is considered unique as they are no
      // longer part of the mount point hash.
      //-----------------------------------------------------------------
      pos = hot_list( ).upper_bound( pos->first );
    }
    //-------------------------------------------------------------------
    // Setup for the return.
    //-------------------------------------------------------------------
    if ( pos != hot_list( ).end( ) )
    {
      //-----------------------------------------------------------------
      // Have a node to scan, so fill in the information for the caller.
      //-----------------------------------------------------------------
      DirInfoRef = pos->second.m_dir_info_ref;
      ExcludeDirs = pos->second.m_exclude_dirs;
    }
    else
    {
      //-----------------------------------------------------------------
      // No more entries so notify the caller of this condition.
      //-----------------------------------------------------------------
      throw no_next( );
    }
    return DirInfoRef;
  }

  //---------------------------------------------------------------------
  /// Start the thread for updating entries that are part of the
  /// collection of hot directories.
  //---------------------------------------------------------------------
  void daemon::
  Start( )
  {
    static bool needs_to_be_started = true;
    if ( needs_to_be_started )
    {
      //-----------------------------------------------------------------
      // Get a thread in which to run this task in the background
      //-----------------------------------------------------------------
      General::TaskThread*      thread = General::ThreadPool::Acquire( );
      daemon*          		task = new daemon;
      //-----------------------------------------------------------------
      /// \todo Need to setup the task thread to be returned to the pool
      ///     once the task is complete.
      //-----------------------------------------------------------------

      //-----------------------------------------------------------------
      // Add the task to the thread and get things running
      //-----------------------------------------------------------------
      thread->AddTask( task );
      needs_to_be_started = false;
    }
  }

  //---------------------------------------------------------------------
  /// Continuously update directories listed in the collection of
  /// hot directories.
  /// After each iteration, pause for specified period of time.
  ///
  /// \todo
  ///     This routine could be made smarter by anticipating when
  ///     the directory will next be updated and suspend till that
  ///     time.
  //---------------------------------------------------------------------
  void daemon::
  operator()( )
  {
    //-------------------------------------------------------------------
    /// Suspend for a period of time
    //-------------------------------------------------------------------
    struct timespec wakeup;
    wakeup.tv_sec = 30;
    wakeup.tv_nsec = 0;
    nanosleep( &wakeup, NULL );

    while( General::AtExit::IsExiting( ) == false )
    {
      //-----------------------------------------------------------------
      // Loop over each element in the hot list
      //-----------------------------------------------------------------
      try
      {
	weak_dir_info_type 			cur;
	General::SharedPtr< DirectoryInfo >	dir;

	while( 1 )
	{
	  dir_container_type		exclude_dirs;
	  diskCache::LogMessage		errors;

	  cur = next( cur, exclude_dirs );

	  if ( ( dir = cur.lock( ) ) )
	  {
	    //-----------------------------------------------------------
	    // Since it is now locked, it is safe to scan
	    //-----------------------------------------------------------
	    bool modified = false;
	    dir->scan( exclude_dirs, errors, modified );
	  }
	}
      }
      catch( const no_next& Exception )
      {
	//---------------------------------------------------------------
	// Have completed the loop around list of hot directories
	//---------------------------------------------------------------
      }
      //-----------------------------------------------------------------
      // Cleanup any entries that need to be removed.
      //-----------------------------------------------------------------
      {
	MutexLock	l( baton( ) );

	for ( hot_directory_container_type::iterator
		hl_cur = hot_list( ).begin( ),
		hl_last = hot_list( ).end( ),
		dead_entry;
	      hl_cur != hl_last; )
	{
	  if ( hl_cur->second.m_dir_info_ref.expired( ) )
	  {
	    //-----------------------------------------------------------
	    // The only reference to the directory information is
	    // within this list of hot directories.
	    //-----------------------------------------------------------
	    dead_entry = hl_cur;
	    ++hl_cur;
	    hot_list( ).erase( dead_entry );
	  }
	  else if ( is_hot( hl_cur->second.m_dir_info_ref ) == false )
	  {
	    //-----------------------------------------------------------
	    // The direcotry has not been recently updated and is
	    // no longer considered hot.
	    //-----------------------------------------------------------
	    std::ostringstream	msg;

	    msg << "Removed " << hl_cur->first
		<< " from the list of hot directories."
	      ;
	    queueLogEntry( msg.str( ),
			   GenericAPI::LogEntryGroup_type::MT_NOTE,
			   0,
			   "daemon::operator()@HotDirectory.cc",
		   JOB_INFO );
	    dead_entry = hl_cur;
	    ++hl_cur;
	    hot_list( ).erase( dead_entry );
	  }
	  else
	  {
	    ++hl_cur;
	  }
	}
      }
      //-----------------------------------------------------------------
      /// Suspend for a period of time
      //-----------------------------------------------------------------
      wakeup.tv_sec = diskCache::HotDirectory::ScanInterval( );
      wakeup.tv_nsec = 0;
      nanosleep( &wakeup, NULL );
    }
  }

} // namespace - anonymous

namespace diskCache
{
  //---------------------------------------------------------------------
  /// A directory that has been modified within this number of
  /// seconds is considered to be hot.
  ///
  /// The default value is 300 seconds ( 5 minutes ).
  //---------------------------------------------------------------------
  mtime_type HotDirectory::m_lower_bound = 300;
  mtime_type HotDirectory::m_scan_interval = 4;

  //---------------------------------------------------------------------
  /// Establish the time representing the lower bounds of what should be
  /// considered hot.
  //---------------------------------------------------------------------
  void HotDirectory::
  HotLowerBound( mtime_type Value )
  {
    m_lower_bound = Value;
  }

  //---------------------------------------------------------------------
  /// Return the time representing the lower bounds of what should be
  /// considered hot.
  //---------------------------------------------------------------------
  mtime_type HotDirectory::
  HotLowerBound( )
  {
    return m_lower_bound;
  }

  //---------------------------------------------------------------------
  /// Checks to see if the directory has been modified within the time
  /// range to be considered hot.
  /// Directories found to be hot are added to the list of hot
  /// directories and are scanned more frequently.
  //---------------------------------------------------------------------
  void HotDirectory::
  RegisterIfHot( weak_dir_info_type Entry,
		 const dir_container_type& ExcludeDirs )
  {
    if ( is_hot( Entry ) )
    {
      //-----------------------------------------------------------------
      // This one needs to be registered
      //-----------------------------------------------------------------
      cond_register( Entry, ExcludeDirs );
    }
  }

  //---------------------------------------------------------------------
  /// Establish the time representing the number of seconds to sleep
  /// between scans of the set of hot directories.
  //---------------------------------------------------------------------
  void HotDirectory::
  ScanInterval( mtime_type Value )
  {
    m_scan_interval = Value;
  }

  //---------------------------------------------------------------------
  /// Remove a directory from the list of hot directories.
  //---------------------------------------------------------------------
  void HotDirectory::
  Unregister( weak_dir_info_type Entry )
  {
    unregister( Entry );
  }
} // namespace - diskCache
