#include "config.h"

#include "general/AtExit.hh"

#include "mountpt.hh"
#include "MountPointScanner.hh"

static bool initialized = false;

namespace diskCache
{
  //-------------------------------------------------------------------------------------
  //
  //-------------------------------------------------------------------------------------
  MutexLock::lock_type	MountPointScanner::SyncRequest::p_request_count_baton
  = MutexLock::Initialize( );
  MountPointScanner::SyncRequest::request_count_type
  MountPointScanner::SyncRequest::p_request_count = 0;

  MountPointScanner::SyncRequest::
  SyncRequest( )
  {
    static const char* caller = "diskCache::MountPointScanner::SyncRequest::SyncRequest";
    MutexLock  l( p_request_count_baton );

    ++p_request_count;

    std::ostringstream msg;

    msg << "There are now " << p_request_count
	<< " pending request(s)"
      ;
    GenericAPI::queueLogEntry( msg.str( ),
			       GenericAPI::LogHTML::MT_DEBUG,
			       30,
			       caller,
			       "IDLE" );
  }

  MountPointScanner::SyncRequest::
  ~SyncRequest( )
  {
    static const char* caller = "diskCache::MountPointScanner::SyncRequest::~SyncRequest";
    MutexLock  l( p_request_count_baton );

    if ( p_request_count > 0 )
    {
      --p_request_count;
    }

    std::ostringstream msg;

    msg << "There are now " << p_request_count
	<< " pending request(s)"
      ;
    GenericAPI::queueLogEntry( msg.str( ),
			       GenericAPI::LogHTML::MT_DEBUG,
			       30,
			       caller,
			       "IDLE" );
  }

  MountPointScanner::SyncRequest::request_count_type MountPointScanner::SyncRequest::
  PendingRequests( )
  {
    static const char* caller = "diskCache::MountPointScanner::SyncRequest::PendingRequests";

    MutexLock  l( p_request_count_baton );

    std::ostringstream msg;

    msg << "There are now " << p_request_count
	<< " pending request(s)"
      ;
    GenericAPI::queueLogEntry( msg.str( ),
			       GenericAPI::LogHTML::MT_DEBUG,
			       50,
			       caller,
			       "IDLE" );

    return p_request_count;
  }

  //-------------------------------------------------------------------------------------
  //
  //-------------------------------------------------------------------------------------
  MutexLock::lock_type		MountPointScanner::m_variable_baton = MutexLock::Initialize( );
  General::ConditionalVariable	MountPointScanner::m_scan_completed;
  MutexLock::lock_type          MountPointScanner::m_active_count_baton = MutexLock::Initialize( );
  MountPointScanner::concurrency_type MountPointScanner::m_active_count = 0;
  MountPointScanner::concurrency_type MountPointScanner::m_concurrency = 10;

  MountPointScanner::scanner_sync_type MountPointScanner::p_scanner_sync = true;
  MountPointScanner::scanner_sync_rw_type::lock_type MountPointScanner::p_scanner_sync_baton
  = General::ReadWriteLock::Initialize( );
  

  void MountPointScanner::
  OnCompletion( int TaskThreadState )
  {
    static const char caller[] = "MountPointScanner::onCompletion";

    if ( ( Results( ).empty( ) == false )
	 && ( General::AtExit::IsExiting( ) == false ) )
    {
      if ( m_stream )
      {
	MutexLock      lock( m_scan_completed.Mutex( ) );

	(*m_stream) << Results( ) << std::endl;
      }
      else
      {
	GenericAPI::queueLogEntry( Results( ),
				   GenericAPI::LogEntryGroup_type::MT_NOTE,
				   10,
				   caller,
				   "SCAN_MOUNTPT" );
      }
    }
    if ( m_controller )
    {
      General::ThreadPool::Relinquish( m_controller );
      m_controller = (General::TaskThread*)NULL;
    }
  }

  void MountPointScanner::
  Scan( const mount_point_container_type& MountPoints,
	std::ostream* Stream )
  {
    //-----------------------------------------------------------
    // Acquire the sync lock
    //-----------------------------------------------------------
    scanner_sync_ro_type	slock( SyncRO( ) );

    if ( initialized == false )
    {
      General::AtExit::Append( on_exit, "diskCache::MountPointScanner::on_exit", 200 );
      initialized = true;
    }
    mount_point_container_type::const_iterator
      cur = MountPoints.begin( ),
      last = MountPoints.end( );

    MountPointScanner::CriticalSection( true );
    while ( ( ( cur != last )
	      || ( Active( ) > 0 ) )
	    && ( General::AtExit::IsExiting( ) == false ) )
    {
      //--------------------------------------------------------------
      // Check to see if scanning should be stopped.
      //--------------------------------------------------------------
      if ( ( SyncRequest::PendingRequests( ) > 0 )
	   && ( cur != last ) )
      {
	cur = last;
      }
      //--------------------------------------------------------------
      // Check to see if there is room to start another scanning thread.
      //--------------------------------------------------------------
      while ( ( Active( ) < Concurrency( ) )
	      && ( cur != last ) )
      {
	start( *cur, Stream );
	++cur; // Advance to the mount point to scan
      }
      //--------------------------------------------------------------
      // Wakeup once a thread has completed.
      //--------------------------------------------------------------
      // MountPointScanner::CriticalSection( true );
      if ( Active( ) )
      {
	MountPointScanner::Wait( );
      }
      // MountPointScanner::CriticalSection( false );
    }
    MountPointScanner::CriticalSection( false );
  }

  void MountPointScanner::
  operator()( )
  {
    try
    {
      m_results = diskCache::MountPointHash::addEntry( m_mount_point );
    }
    catch( std::exception& Exception )
    {
      m_results = Exception.what( );
    }
    catch( ... )
    {
      m_results = "MountPointScanner received an unknown exception";
    }
    CriticalSection( true );
    m_scan_completed.Broadcast( );
    CriticalSection( false );
  }

  void MountPointScanner::
  on_exit( )
  {
    //----------------------------------------------------------------
    // Get the attention of the scanning daemon
    //----------------------------------------------------------------
    m_scan_completed.Broadcast( );
    MountPointScanner::CriticalSection( true );
    while ( Active( ) > 0 )
    {
      //--------------------------------------------------------------
      // Wakeup once a thread has completed.
      //--------------------------------------------------------------
      // MountPointScanner::CriticalSection( true );
      if ( Active( ) )
      {
	MountPointScanner::Wait( );
      }
      // MountPointScanner::CriticalSection( false );
    }
    MountPointScanner::CriticalSection( false );
    
  }

  inline void MountPointScanner::
  start( const std::string& MountPoint, std::ostream* Stream )
  {
    if ( General::AtExit::IsExiting( ) == true )
    {
      return;
    }

    General::TaskThread*   thread = General::ThreadPool::Acquire( );
    MountPointScanner*     task = new MountPointScanner( MountPoint, Stream, thread );

    try
    {
      thread->AddTask( task );
    }
    catch( ... )
    {
      throw;
    }
  }

}
